@@ -43,7 +43,6 @@
 #include "axel.h"
 #include "sleep.h"
 
-static char *axel_strrstr(char *haystack, char *needle);
 static void *search_speedtest(void *r);
 static int search_sortlist_qsort(const void *a, const void *b);
 
@@ -63,11 +62,10 @@ main(int argc, char *argv[])
 	conf_init(conf);
 	ssl_init(conf);
 
-	res = malloc(sizeof(search_t) * (conf->search_amount + 1));
+	res = calloc(conf->search_amount + 1, sizeof(search_t));
 	if (!res)
 		goto out;
 
-	memset(res, 0, sizeof(search_t) * (conf->search_amount + 1));
 	res->conf = conf;
 
 	i = search_makelist(res, argv[1]);
@@ -94,91 +92,103 @@ out:
 #endif
 
 int
-search_makelist(search_t *results, char *url)
+search_makelist(search_t *results, char *orig_url)
 {
-	int i, size = 8192, j = 0;
-	char *s, *s1, *s3;
+	int size = 8192;
 	conn_t conn[1];
 	double t;
+	const char *start, *end;
 
 	memset(conn, 0, sizeof(conn_t));
 
 	conn->conf = results->conf;
-	t = gettime();
-	if (!conn_set(conn, url) || !conn_init(conn) || !conn_info(conn))
+	t = axel_gettime();
+	if (!conn_set(conn, orig_url) || !conn_init(conn) || !conn_info(conn))
 		return -1;
 
-	strncpy(results[0].url, url, sizeof(results[0].url) - 1);
-	results[0].url[sizeof(results[0].url) - 1] = '\0';
-	results[0].speed = 1 + 1000 * (gettime() - t);
+	size_t orig_len = strlcpy(results[0].url, orig_url,
+				  sizeof(results[0].url));
+	results[0].speed = 1 + 1000 * (axel_gettime() - t);
 	results[0].size = conn->size;
+	int nresults = 1;
 
-	s = malloc(size);
+	char *s = malloc(size);
 	if (!s)
-		return -1;
+		return 1;
 
-	sprintf(s, "http://www.filesearching.com/cgi-bin/s?q=%s&w=a&l=en&"
-		"t=f&e=on&m=%i&o=n&s1=%lld&s2=%lld&x=15&y=15",
-		conn->file, results->conf->search_amount,
-		conn->size, conn->size);
+	/* TODO improve matches */
+	snprintf(s, size, "http://www.filesearching.com/cgi-bin/s?"
+		 "w=a&" /* TODO describe */
+		 "x=15&y=15&" /* TODO describe */
+		 /* Size in bytes:   */ "s=on&"
+		 /* Exact search:    */ "e=on&"
+		 /* Language:        */ "l=en&"
+		 /* Search Type:     */ "t=f&"
+		 /* Sorting:         */ "o=n&"
+		 /* Filename:        */ "q=%s&"
+		 /* Num. of results: */ "m=%i&"
+		 /* Size (min/max):  */ "s1=%lld&s2=%lld",
+		 conn->file, results->conf->search_amount,
+		 conn->size, conn->size);
 
 	conn_disconnect(conn);
 	memset(conn, 0, sizeof(conn_t));
 	conn->conf = results->conf;
 
-	if (!conn_set(conn, s) || !conn_setup(conn) || !conn_exec(conn)) {
-		free(s);
-		return 1;
-	}
+	if (!conn_set(conn, s) || !conn_setup(conn) || !conn_exec(conn))
+		goto done;
 
-	while ((i = tcp_read(conn->tcp, s + j, size - j)) > 0) {
-		j += i;
-		if (j + 10 >= size) {
-			size *= 2;
-			char *tmp = realloc(s, size);
-			if (!tmp) {
-				free(s);
-				return 1;
+	{
+		int j = 0;
+
+		for (int i; (i = tcp_read(conn->tcp, s + j, size - j)) > 0;) {
+			j += i;
+			if (j + 10 >= size) {
+				size *= 2;
+				char *tmp = realloc(s, size);
+				if (!tmp)
+					goto done;
+				s = tmp;
+				memset(s + size / 2, 0, size / 2);
 			}
-			s = tmp;
-			memset(s + size / 2, 0, size / 2);
 		}
+		s[j] = '\0';
 	}
-	s[j] = '\0';
 
 	conn_disconnect(conn);
 
-	s1 = strstr(s, "<pre class=list");
-	s1 = strchr(s1, '\n') + 1;
-	if (strstr(s1, "</pre>") == NULL) {
-		/* Incomplete list */
-		free(s);
-		return 1;
-	}
-	for (i = 1;
-	     strncmp(s1, "</pre>", 6) && i < results->conf->search_amount && *s1;
-         i++) {
-		s3 = strchr(s1, '\n');
-		*s3 = 0;
-		char *s2 = axel_strrstr(s1, "<a href=") + 8;
-		*s3 = '\n';
-		s3 = strchr(s2, ' ');
-		*s3 = 0;
-		if (strcmp(results[0].url, s2)) {
-			strncpy(results[i].url, s2, sizeof(results[i].url) - 1);
-			results[i].size = results[0].size;
-			results[i].conf = results->conf;
-		} else {
-			/* The original URL might show up */
-			i--;
-		}
-		for (s1 = s3; *s1 != '\n'; s1++) ;
-		s1++;
+	start = strstr(s, "<pre class=list");
+	if (!start)
+		goto done;
+	end = strstr(start, "</pre>");
+	/* Incomplete list */
+	if (!end)
+		goto done;
+
+	for (const char *url, *eol;
+	     start < end && nresults < results->conf->search_amount;
+	     start = eol + 1) {
+		eol = strchr(start, '\n');
+		if (eol > end || !eol)
+			eol = end;
+		do {
+			url = start;
+			start = strstr(start, "<a href=") + 8;
+		} while (start < eol);
+
+		/* Check it isn't the original URL */
+		if (!strncmp(url, orig_url, orig_len))
+			continue;
+
+		strlcpy(results[nresults].url, url, sizeof(results[0].url));
+		results[nresults].size = results[0].size;
+		results[nresults].conf = results->conf;
+		++nresults;
 	}
 
+done:
 	free(s);
-
-	return i;
+	return nresults;
 }
 
 enum {
@@ -208,7 +218,7 @@ search_getspeeds(search_t *results, int count)
 		for (int i = 0; i < count; i++) {
 			switch (results[i].speed) {
 			case SPEED_ACTIVE:
-				if (gettime() < results[i].speed_start_time
+				if (axel_gettime() < results[i].speed_start_time
 				    + results->conf->search_timeout)
 					continue; // not timed out yet
 				pthread_cancel(*results[i].speed_thread);
@@ -221,7 +231,7 @@ search_getspeeds(search_t *results, int count)
 				if (running >= results->conf->search_threads)
 					continue; // running too many, skip
 				results[i].speed = SPEED_ACTIVE;
-				results[i].speed_start_time = gettime();
+				results[i].speed_start_time = axel_gettime();
 				if (pthread_create(results[i].speed_thread,
 						   NULL, search_speedtest,
 						   &results[i]) == 0)
@@ -254,6 +264,7 @@ search_getspeeds(search_t *results, int count)
 	return correct;
 }
 
+static
 void *
 search_speedtest(void *r)
 {
@@ -272,7 +283,7 @@ search_speedtest(void *r)
 	    && conn_info(conn)
 	    && conn->size == results->size)
 		/* Add one because it mustn't be zero */
-		results->speed = 1 + 1000 * (gettime() - results->speed_start_time);
+		results->speed = 1 + 1000 * (axel_gettime() - results->speed_start_time);
 	else
 		results->speed = SPEED_FAILED;
 
@@ -281,19 +292,6 @@ search_speedtest(void *r)
 	return NULL;
 }
 
-char *
-axel_strrstr(char *haystack, char *needle)
-{
-	int i, j;
-
-	for (i = strlen(haystack) - strlen(needle); i > 0; i--) {
-		for (j = 0; needle[j] && haystack[i + j] == needle[j]; j++) ;
-		if (!needle[j])
-			return haystack + i;
-	}
-
-	return NULL;
-}
 
 void
 search_sortlist(search_t *results, int count)
@@ -301,6 +299,7 @@ search_sortlist(search_t *results, int count)
 	qsort(results, count, sizeof(search_t), search_sortlist_qsort);
 }
 
+static
 int
 search_sortlist_qsort(const void *a, const void *b)
 {
