@@ -46,36 +46,57 @@
 #include <sys/socket.h>
 #include <netdb.h>
 #include <sys/ioctl.h>
+#include <netinet/tcp.h>
 
+#ifndef TCP_FASTOPEN_CONNECT
+#ifdef __linux__
+#define TCP_FASTOPEN_CONNECT 30
+#else /* __linux__ */
+#define TCP_FASTOPEN_CONNECT 0
+#endif /* __linux__ */
+#endif /* !TCP_FASTOPEN_CONNECT */
 
-static void
-tcp_error(char *buffer, char *hostname, int port, const char *reason)
+/*
+ * Check if the given hostname is ipv6 literal
+ * Returns 1 if true and 0 if false
+ */
+int is_ipv6_addr(const char *hostname) {
+	char buf[16]; /* Max buff size needed for inet_pton() */
+
+	return hostname && 1 == inet_pton(AF_INET6, hostname, buf);
+}
+
+inline static void
+tcp_error(char *hostname, int port, const char *reason)
 {
-	sprintf(buffer, _("Unable to connect to server %s:%i: %s\n"),
+	fprintf(stderr, _("Unable to connect to server %s:%i: %s\n"),
 		hostname, port, reason);
 }
 
 /* Get a TCP connection */
 int
 tcp_connect(tcp_t *tcp, char *hostname, int port, int secure, char *local_if,
-	    char *message, unsigned io_timeout)
+	    unsigned io_timeout)
 {
 	struct sockaddr_in local_addr;
-	const int portstr_len = 10;
-	char portstr[portstr_len];
+	char portstr[10];
 	struct addrinfo ai_hints;
 	struct addrinfo *gai_results, *gai_result;
 	int ret;
 	int sock_fd = -1;
 
 	memset(&local_addr, 0, sizeof(local_addr));
-	if (tcp->ai_family == AF_INET && local_if && *local_if) {
-		local_addr.sin_family = AF_INET;
-		local_addr.sin_port = 0;
-		local_addr.sin_addr.s_addr = inet_addr(local_if);
+	if (local_if) {
+		if (!*local_if || tcp->ai_family != AF_INET) {
+			local_if = NULL;
+		} else {
+			local_addr.sin_family = AF_INET;
+			local_addr.sin_port = 0;
+			local_addr.sin_addr.s_addr = inet_addr(local_if);
+		}
 	}
 
-	snprintf(portstr, portstr_len, "%d", port);
+	snprintf(portstr, sizeof(portstr), "%d", port);
 
 	memset(&ai_hints, 0, sizeof(ai_hints));
 	ai_hints.ai_family = tcp->ai_family;
@@ -85,71 +106,75 @@ tcp_connect(tcp_t *tcp, char *hostname, int port, int secure, char *local_if,
 
 	ret = getaddrinfo(hostname, portstr, &ai_hints, &gai_results);
 	if (ret != 0) {
-		tcp_error(message, hostname, port, gai_strerror(ret));
+		tcp_error(hostname, port, gai_strerror(ret));
 		return -1;
 	}
 
 	gai_result = gai_results;
-	sock_fd = -1;
-	while ((sock_fd == -1) && (gai_result != NULL)) {
+	do {
+		int tcp_fastopen = -1;
 
+		if (sock_fd != -1) {
+			close(sock_fd);
+			sock_fd = -1;
+		}
 		sock_fd = socket(gai_result->ai_family,
 				 gai_result->ai_socktype,
 				 gai_result->ai_protocol);
+		if (sock_fd == -1)
+			continue;
 
-		if (sock_fd != -1) {
-
-			if (gai_result->ai_family == AF_INET) {
-				if (local_if && *local_if) {
-					ret = bind(sock_fd,
-						   (struct sockaddr *)
-						   &local_addr,
-						   sizeof(local_addr));
-					if (ret == -1) {
-						close(sock_fd);
-						sock_fd = -1;
-						gai_result =
-						    gai_result->ai_next;
-					}
-				}
-			}
-
-			if (sock_fd != -1) {
-				struct timeval tout = { .tv_sec  = io_timeout };
-				/* Set O_NONBLOCK so we can timeout */
-				if (io_timeout)
-					fcntl(sock_fd, F_SETFL, O_NONBLOCK);
-				ret = connect(sock_fd, gai_result->ai_addr,
-					      gai_result->ai_addrlen);
-				/* Wait for the connection */
-				if (ret == -1 && errno == EINPROGRESS) {
-					fd_set fdset;
-					FD_ZERO(&fdset);
-					FD_SET(sock_fd, &fdset);
-					ret = select(sock_fd + 1,
-						     NULL, &fdset, NULL,
-						     &tout);
-				}
-				if (ret == -1) {
-					close(sock_fd);
-					sock_fd = -1;
-					gai_result = gai_result->ai_next;
-				} else {
-					fcntl(sock_fd, F_SETFL, 0);
-				}
-			}
+		if (local_if && gai_result->ai_family == AF_INET) {
+			bind(sock_fd, (struct sockaddr *)&local_addr,
+			     sizeof(local_addr));
+			/* FIXME report errors */
 		}
-	}
+
+		if (TCP_FASTOPEN_CONNECT) {
+			tcp_fastopen = setsockopt(sock_fd, IPPROTO_TCP,
+						  TCP_FASTOPEN_CONNECT,
+						  NULL, 0);
+		} else if (io_timeout) {
+			/* Set O_NONBLOCK so we can timeout */
+			fcntl(sock_fd, F_SETFL, O_NONBLOCK);
+		}
+		ret = connect(sock_fd, gai_result->ai_addr,
+			      gai_result->ai_addrlen);
+
+		/* Already connected maybe? */
+		if (ret != -1)
+			break;
+
+		if (errno != EINPROGRESS)
+			continue;
+
+		/* With TFO we must assume success */
+		if (tcp_fastopen != -1)
+			break;
+
+		/* Wait for the connection */
+		fd_set fdset;
+		FD_ZERO(&fdset);
+		FD_SET(sock_fd, &fdset);
+		struct timeval tout = { .tv_sec  = io_timeout };
+		ret = select(sock_fd + 1, NULL, &fdset, NULL, &tout);
+		/* Success? */
+		if (ret != -1)
+			break;
+	} while ((gai_result = gai_result->ai_next));
 
 	freeaddrinfo(gai_results);
 
 	if (sock_fd == -1) {
-		tcp_error(message, hostname, port, strerror(errno));
+		tcp_error(hostname, port, strerror(errno));
 		return -1;
 	}
+
+	fcntl(sock_fd, F_SETFL, 0);
+
 #ifdef HAVE_SSL
 	if (secure) {
-		tcp->ssl = ssl_connect(sock_fd, hostname, message);
+		tcp->ssl = ssl_connect(sock_fd, hostname);
 		if (tcp->ssl == NULL) {
 			close(sock_fd);
 			return -1;
@@ -166,7 +191,7 @@ tcp_connect(tcp_t *tcp, char *hostname, int port, int secure, char *local_if,
 	return 1;
 }
 
-int
+ssize_t
 tcp_read(tcp_t *tcp, void *buffer, int size)
 {
 #ifdef HAVE_SSL
@@ -177,7 +202,7 @@ tcp_read(tcp_t *tcp, void *buffer, int size)
 		return read(tcp->fd, buffer, size);
 }
 
-int
+ssize_t
 tcp_write(tcp_t *tcp, void *buffer, int size)
 {
 #ifdef HAVE_SSL
@@ -204,7 +229,7 @@ tcp_close(tcp_t *tcp)
 }
 
 int
-get_if_ip(char *iface, char *ip)
+get_if_ip(char *dst, size_t len, const char *iface)
 {
 	struct ifreq ifr;
 	int ret, fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
@@ -214,14 +239,13 @@ get_if_ip(char *iface, char *ip)
 
 	memset(&ifr, 0, sizeof(struct ifreq));
 
-	strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name) - 1);
-	ifr.ifr_name[sizeof(ifr.ifr_name) - 1] = '\0';
+	strlcpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));
 	ifr.ifr_addr.sa_family = AF_INET;
 
 	ret = !ioctl(fd, SIOCGIFADDR, &ifr);
 	if (ret) {
 		struct sockaddr_in *x = (struct sockaddr_in *)&ifr.ifr_addr;
-		strcpy(ip, inet_ntoa(x->sin_addr));
+		strlcpy(dst, inet_ntoa(x->sin_addr), len);
 	}
 	close(fd);
 
