@@ -48,13 +48,14 @@
 
 
 static void stop(int signal);
-static char *size_human(long long int value);
-static char *time_human(int value);
+static char *size_human(char *dst, size_t len, size_t value);
+static char *time_human(char *dst, size_t len, unsigned int value);
 static void print_commas(long long int bytes_done);
 static void print_alternate_output(axel_t *axel);
-static void print_help();
-static void print_version();
-static int get_term_width();
+static void print_progress(size_t cur, size_t prev, size_t total, double kbps);
+static void print_help(void);
+static void print_version(void);
+static int get_term_width(void);
 
 int run = 1;
 
@@ -87,20 +88,19 @@ static struct option axel_options[] = {
 };
 #endif
 
-/* For returning string values from functions */
-static char string[MAX_STRING + 3];
-
 int
 main(int argc, char *argv[])
 {
-	char fn[MAX_STRING] = "";
+	char fn[MAX_STRING];
 	int do_search = 0;
 	search_t *search;
 	conf_t conf[1];
 	axel_t *axel;
-	int i, j, cur_head = 0, ret = 1;
+	int j, cur_head = 0, ret = 1;
 	char *s;
 
+	fn[0] = 0;
+
 /* Set up internationalization (i18n) */
 #ifdef ENABLE_NLS
 	setlocale(LC_ALL, "");
@@ -116,22 +116,20 @@ main(int argc, char *argv[])
 
 	j = -1;
 	while (1) {
-		int option;
-
-		option =
-		    getopt_long(argc, argv, "s:n:o:S::46NqvhVakcH:U:T:",
-				axel_options, NULL);
+		int option = getopt_long(argc, argv,
+					 "s:n:o:S::46NqvhVakcH:U:T:",
+					 axel_options, NULL);
 		if (option == -1)
 			break;
 
 		switch (option) {
 		case 'U':
-			strncpy(conf->user_agent, optarg,
-				sizeof(conf->user_agent) - 1);
+			conf_hdr_make(conf->add_header[HDR_USER_AGENT],
+				      "User-Agent", optarg);
 			break;
 		case 'H':
-			strncpy(conf->add_header[cur_head++], optarg,
-				sizeof(conf->add_header[cur_head - 1]) - 1);
+			strlcpy(conf->add_header[cur_head++], optarg,
+				sizeof(conf->add_header[0]));
 			break;
 		case 's':
 			if (!sscanf(optarg, "%i", &conf->max_speed)) {
@@ -152,8 +150,7 @@ main(int argc, char *argv[])
 			}
 			break;
 		case 'o':
-			strncpy(fn, optarg, sizeof(fn) - 1);
-			fn[sizeof(fn) - 1] = '\0';
+			strlcpy(fn, optarg, sizeof(fn));
 			break;
 		case 'S':
 			do_search = 1;
@@ -252,7 +249,7 @@ main(int argc, char *argv[])
 		s = argv[optind];
 		if (strlen(s) > MAX_STRING) {
 			fprintf(stderr,
-				_("Can't handle URLs of length over %d\n"),
+				_("Can't handle URLs of length over %zu\n"),
 				MAX_STRING);
 			goto free_conf;
 		}
@@ -260,15 +257,14 @@ main(int argc, char *argv[])
 
 	printf(_("Initializing download: %s\n"), s);
 	if (do_search) {
-		search = malloc(sizeof(search_t) * (conf->search_amount + 1));
+		search = calloc(conf->search_amount + 1, sizeof(search_t));
 		if (!search)
 			goto free_conf;
 
-		memset(search, 0, sizeof(search_t) * (conf->search_amount + 1));
 		search[0].conf = conf;
 		if (conf->verbose)
 			printf(_("Doing search...\n"));
-		i = search_makelist(search, s);
+		int i = search_makelist(search, s);
 		if (i < 0) {
 			fprintf(stderr, _("File not found\n"));
 			goto free_conf;
@@ -305,14 +301,13 @@ main(int argc, char *argv[])
 			goto close_axel;
 		}
 	} else {
-		search = malloc(sizeof(search_t) * (argc - optind));
+		search = calloc(argc - optind, sizeof(search_t));
 		if (!search)
 			goto free_conf;
 
-		memset(search, 0, sizeof(search_t) * (argc - optind));
-		for (i = 0; i < (argc - optind); i++)
-			strncpy(search[i].url, argv[optind + i],
-				sizeof(search[i].url) - 1);
+		for (int i = 0; i < argc - optind; i++)
+			strlcpy(search[i].url, argv[optind + i],
+				sizeof(search[i].url));
 		axel = axel_new(conf, argc - optind, search);
 		free(search);
 		if (!axel || axel->ready == -1) {
@@ -325,6 +320,7 @@ main(int argc, char *argv[])
 		free(s);
 	}
 
+	/* Check if a file name has been specified */
 	if (*fn) {
 		struct stat buf;
 
@@ -334,8 +330,7 @@ main(int argc, char *argv[])
 				size_t axelfnlen = strlen(axel->filename);
 
 				if (fnlen + 1 + axelfnlen + 1 > MAX_STRING) {
-					fprintf(stderr,
-						_("Filename too long!\n"));
+					fprintf(stderr, _("Filename too long!\n"));
 					goto close_axel;
 				}
 
@@ -345,33 +340,34 @@ main(int argc, char *argv[])
 				fn[fnlen + 1 + axelfnlen] = '\0';
 			}
 		}
-		sprintf(string, "%s.st", fn);
-		if (access(fn, F_OK) == 0 && access(string, F_OK) != 0) {
+		char statefn[MAX_STRING + 3];
+		snprintf(statefn, sizeof(statefn), "%s.st", fn);
+		if (access(fn, F_OK) == 0 && access(statefn, F_OK) != 0) {
 			fprintf(stderr, _("No state file, cannot resume!\n"));
 			goto close_axel;
 		}
-		if (access(string, F_OK) == 0 && access(fn, F_OK) != 0) {
+		if (access(statefn, F_OK) == 0 && access(fn, F_OK) != 0) {
 			printf(_("State file found, but no downloaded data. Starting from scratch.\n"));
-			unlink(string);
+			unlink(statefn);
 		}
-		strcpy(axel->filename, fn);
+		strlcpy(axel->filename, fn, sizeof(axel->filename));
 	} else {
 		/* Local file existence check */
-		i = 0;
 		s = axel->filename + strlen(axel->filename);
-		while (1) {
-			sprintf(string, "%s.st", axel->filename);
-			if (access(axel->filename, F_OK) == 0) {
-				if (axel->conn[0].supported) {
-					if (access(string, F_OK) == 0)
+		for (int i = 0; 1; i++) {
+			char statefn[MAX_STRING + 3];
+			snprintf(statefn, sizeof(statefn), "%s.st",
+				 axel->filename);
+
+			int f_exists = !access(axel->filename, F_OK);
+			int st_exists = !access(statefn, F_OK);
+			if (f_exists) {
+				if (axel->conn[0].supported && st_exists)
 						break;
-				}
-			} else {
-				if (access(string, F_OK))
-					break;
-			}
-			sprintf(s, ".%i", i);
-			i++;
+			} else if (!st_exists)
+				break;
+			snprintf(s, axel->filename + sizeof(axel->filename) - s,
+				 ".%i", i);
 		}
 	}
 
@@ -389,6 +385,7 @@ main(int argc, char *argv[])
 		if (axel->bytes_done > 0) {	/* Print first dots if resuming */
 			putchar('\n');
 			print_commas(axel->bytes_done);
+			fflush(stdout);
 		}
 	}
 	axel->start_byte = axel->bytes_done;
@@ -398,7 +395,7 @@ main(int argc, char *argv[])
 	signal(SIGTERM, stop);
 
 	while (!axel->ready && run) {
-		long long int prev;
+		size_t prev;
 
 		prev = axel->bytes_done;
 		axel_do(axel);
@@ -406,46 +403,16 @@ main(int argc, char *argv[])
 		if (conf->alternate_output) {
 			if (!axel->message && prev != axel->bytes_done)
 				print_alternate_output(axel);
-		} else {
-			/* The infamous wget-like 'interface'.. ;) */
-			long long int done =
-			    (axel->bytes_done / 1024) - (prev / 1024);
-			if (done && conf->verbose > -1) {
-				for (i = 0; i < done; i++) {
-					i += (prev / 1024);
-					if ((i % 50) == 0) {
-						if (prev >= 1024)
-							printf("  [%6.1fKB/s]",
-							       (double)axel->bytes_per_second /
-							       1024);
-						if (axel->size == LLONG_MAX)
-							printf("\n[ N/A]  ");
-						else if (axel->size < 10240000)
-							printf("\n[%3lld%%]  ",
-							       min(100,
-								   102400 * i /
-								   axel->size));
-						else
-							printf("\n[%3lld%%]  ",
-							       min(100,
-								   i /
-								   (axel->size /
-								    102400)));
-					} else if ((i % 10) == 0) {
-						putchar(' ');
-					}
-					putchar('.');
-					i -= (prev / 1024);
-				}
-				fflush(stdout);
-			}
+		} else if (conf->verbose > -1) {
+			print_progress(axel->bytes_done, prev, axel->size,
+				       (double)axel->bytes_per_second / 1024);
 		}
 
 		if (axel->message) {
 			if (conf->alternate_output == 1) {
 				/* clreol-simulation */
 				putchar('\r');
-				for (i = get_term_width(); i > 0; i--)
+				for (int i = get_term_width(); i > 0; i--)
 					putchar(' ');
 				putchar('\r');
 			} else {
@@ -461,14 +428,14 @@ main(int argc, char *argv[])
 		} else if (axel->ready) {
 			putchar('\n');
 		}
+		fflush(stdout);
 	}
 
-	strcpy(string + MAX_STRING / 2,
-	       size_human(axel->bytes_done - axel->start_byte));
+	char hsize[MAX_STRING / 2], htime[MAX_STRING / 2];
+	time_human(htime, sizeof(htime), axel_gettime() - axel->start_time);
+	size_human(hsize, sizeof(hsize), axel->bytes_done - axel->start_byte);
 
-	printf(_("\nDownloaded %s in %s. (%.2f KB/s)\n"),
-	       string + MAX_STRING / 2,
-	       time_human(gettime() - axel->start_time),
+	printf(_("\nDownloaded %s in %s. (%.2f KB/s)\n"), hsize, htime,
 	       (double)axel->bytes_per_second / 1024);
 
 	ret = axel->ready ? 0 : 2;
@@ -489,39 +456,51 @@ stop(int signal)
 	run = 0;
 }
 
+/**
+ * Integer base-2 logarithm.
+ */
+static inline
+unsigned
+log2i(unsigned long long x)
+{
+	return x ? sizeof(x) * 8 - 1 - __builtin_clzll(x) : 0;
+}
+
 /* Convert a number of bytes to a human-readable form */
 char *
-size_human(long long int value)
+size_human(char *dst, size_t len, size_t value)
 {
-	if (value < 1024)
-		sprintf(string, _("%lld byte"), value);
-	else if (value < 1024 * 1024)
-		sprintf(string, _("%.1f Kilobyte"), (float)value / 1024);
-	else if (value < 1024 * 1024 * 1024)
-		sprintf(string, _("%.1f Megabyte"),
-			(float)value / (1024 * 1024));
-	else
-		sprintf(string, _("%.1f Gigabyte"),
-			(float)value / (1024 * 1024 * 1024));
+	double fval = (double)value;
+	const char * const oname[] = {
+		"", _("Kilo"), _("Mega"), _("Giga"), _("Tera"),
+	};
+	const unsigned int order = min(sizeof(oname) / sizeof(oname[0]) - 1,
+				       log2i(fval) / 10);
 
-	return string;
+	fval /= (double)(1 << order * 10);
+	int ret = snprintf(dst, len, _("%g %sbyte(s)"), fval, oname[order]);
+	return ret < 0 ? NULL : dst;
 }
 
 /* Convert a number of seconds to a human-readable form */
 char *
-time_human(int value)
+time_human(char *dst, size_t len, unsigned int value)
 {
-	if (value == 1)
-		sprintf(string, _("%i second"), value);
-	else if (value < 60)
-		sprintf(string, _("%i seconds"), value);
-	else if (value < 3600)
-		sprintf(string, _("%i:%02i minute(s)"), value / 60, value % 60);
-	else
-		sprintf(string, _("%i:%02i:%02i hour(s)"), value / 3600,
-			(value / 60) % 60, value % 60);
+	unsigned int hh, mm, ss;
 
-	return string;
+	ss = value % 60;
+	mm = value / 60 % 60;
+	hh = value / 3600;
+
+	int ret;
+	if (hh)
+		ret = snprintf(dst, len, _("%i:%02i:%02i hour(s)"), hh, mm, ss);
+	else if (mm)
+		ret = snprintf(dst, len, _("%i:%02i minute(s)"), mm, ss);
+	else
+		ret = snprintf(dst, len, _("%i second(s)"), ss);
+
+	return ret < 0 ? NULL : dst;
 }
 
 /* Part of the infamous wget-like interface. Just put it in a function
@@ -540,32 +519,60 @@ print_commas(long long int bytes_done)
 			putchar(' ');
 		putchar(',');
 	}
-	fflush(stdout);
 }
 
+
+/**
+ * The infamous wget-like 'interface'.. ;)
+ */
+static
+void
+print_progress(size_t cur, size_t prev, size_t total, double kbps)
+{
+	prev /= 1024;
+	cur /= 1024;
+
+	bool print_speed = prev > 0;
+	for (size_t i = prev; i < cur; i++) {
+		if (i % 50 == 0) {
+			if (print_speed)
+				printf("  [%6.1fKB/s]", kbps);
+
+			if (total == LLONG_MAX)
+				printf("\n[ N/A]  ");
+			else
+				printf("\n[%3zu%%]  ",
+				       min(100U, 102400 * i / total));
+		} else if (i % 10 == 0) {
+			putchar(' ');
+		}
+		putchar('.');
+	}
+}
+
+
 static void
 print_alternate_output_progress(axel_t *axel, char *progress, int width,
 				long long int done, long long int total,
 				double now)
 {
+	if (!width)
+		width = 1;
+	if (!total)
+		total = 1;
 	for (int i = 0; i < axel->conf->num_connections; i++) {
-		int offset = ((double)(axel->conn[i].currentbyte) / (total + 1)
-			      * (width + 1));
+		int offset = axel->conn[i].currentbyte * width / total;
 
 		if (axel->conn[i].currentbyte < axel->conn[i].lastbyte) {
 			if (now <= axel->conn[i].last_transfer
 				   + axel->conf->connection_timeout / 2) {
-				if (i < 10)
-					progress[offset] = i + '0';
-				else
-					progress[offset] = i - 10 + 'A';
+				progress[offset] = i
+					+ (i < 10 ? '0' : ('A' - 10));
 			} else
 				progress[offset] = '#';
 		}
-		int end = ((double)(axel->conn[i].lastbyte) / (total + 1)
-			   * (width + 1));
-		for (int j = offset + 1; j < end; j++)
-			progress[j] = ' ';
+		memset(progress + offset + 1, ' ',
+		       max(0, axel->conn[i].lastbyte * width / total - offset - 1));
 	}
 
 	progress[width] = '\0';
@@ -578,7 +585,7 @@ print_alternate_output(axel_t *axel)
 {
 	long long int done = axel->bytes_done;
 	long long int total = axel->size;
-	double now = gettime();
+	double now = axel_gettime();
 	int width = get_term_width();
 	char *progress;
 
@@ -630,13 +637,11 @@ print_alternate_output(axel_t *axel)
 			printf(" [%02d:%02d]", minutes, seconds);
 	}
 
-	fflush(stdout);
-
 	free(progress);
 }
 
 static int
-get_term_width()
+get_term_width(void)
 {
 	struct winsize w;
 
@@ -645,7 +650,7 @@ get_term_width()
 }
 
 void
-print_help()
+print_help(void)
 {
 #ifdef NOGETOPTLONG
 	printf(_("Usage: axel [options] url1 [url2] [url...]\n"
@@ -696,7 +701,7 @@ print_help()
 }
 
 void
-print_version()
+print_version(void)
 {
 	printf(_("Axel version %s (%s)\n"), VERSION, ARCH);
 	printf("\nCopyright 2001-2007 Wilmer van der Gaast,\n"
